from abc import ABC, abstractmethod
from typing import Any, Type, Union, Annotated, get_origin, get_args, List, Optional
from uuid import UUID as _UUID

from bson import Binary
from bson import ObjectId as _ObjectId
from pydantic import GetCoreSchemaHandler, BaseModel
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from pydantic.functional_serializers import PlainSerializer


def is_optional(field):
    return get_origin(field) is Union and \
           type(None) in get_args(field)


class ObjectIdPydanticAnnotation:
    @classmethod
    def validate_object_id(cls, v: Any, handler) -> _ObjectId:
        if isinstance(v, _ObjectId):
            return v

        s = handler(v)
        if _ObjectId.is_valid(s):
            return _ObjectId(s)
        else:
            raise ValueError("Invalid ObjectId")

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, _handler) -> core_schema.CoreSchema:
        assert source_type is _ObjectId or (is_optional(source_type) and _ObjectId in get_args(source_type))
        return core_schema.no_info_wrap_validator_function(
            cls.validate_object_id,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


ObjectId = Annotated[_ObjectId, ObjectIdPydanticAnnotation]


class Uuid(_UUID):
    @classmethod
    def validate_uuid(cls, v: Any) -> _UUID:
        if isinstance(v, _UUID): return v
        if isinstance(v, str): return _UUID(v)
        if isinstance(v, Binary): return v.as_uuid()

    # def to_binary(self):
    #     return Binary.from_uuid(self)

    @classmethod
    def from_uuid(cls, payload) -> Binary:
        if isinstance(payload, Binary): return payload
        return Binary.from_uuid(payload)

    @classmethod
    def __get_pydantic_core_schema__(
            cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate_uuid,
            core_schema.uuid_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=cls.from_uuid
            ),
        )

# add json plain serializer to Uuid
# Uuid = Annotated[
#     Uuid, PlainSerializer(lambda x: str(x), return_type=str, when_used='json')
# ]

MaybeUuid = Union[Uuid, _UUID, Binary]


def to_binary_uuid(uuid_: Union[MaybeUuid, List[MaybeUuid]]):
    if isinstance(uuid_, list): return [to_binary_uuid(u) for u in uuid_]
    if isinstance(uuid_, Binary): return uuid_
    return Binary.from_uuid(uuid_)


def serialize_object_ids_in_dict(mey_be_dict: Any):
    if not isinstance(mey_be_dict, dict): return
    for key, val in mey_be_dict.items():
        if isinstance(val, _ObjectId):
            mey_be_dict[key] = str(val)
        elif isinstance(val, dict):
            serialize_object_ids_in_dict(val)
        elif isinstance(val, list):
            for item in val:
                serialize_object_ids_in_dict(item)


class DbSettingsABC(ABC):

    @property
    @abstractmethod
    def qdrant_host(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def qdrant_port(self) -> Optional[int]: ...

    @property
    @abstractmethod
    def mongo_db_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def redis_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_write_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_read_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_uri(self) -> Optional[str]: ...


class QueryBase(BaseModel):
    offset: Optional[int] = None
    limit: Optional[int] = None
    filters: Optional[Union[dict, List[Any]]] = None
    order_by: Optional[Any] = None

    id: Optional[Union[ObjectId, Uuid, str]] = None
    ids: Optional[List[Union[ObjectId, Uuid, str]]] = None

    @classmethod
    def loop(cls, func, *args, **kwargs) -> List[Any]:
        page = 0
        page_size = 100
        res = []
        has_more = True
        base_query: QueryBase = kwargs.get("query")
        if 'query' in kwargs: del kwargs['query']
        base_query_dict: dict = base_query.model_dump() if base_query is not None else {}
        if 'offset' in base_query_dict: del base_query_dict['offset']
        if 'limit' in base_query_dict: del base_query_dict['limit']
        while page < 10000 and has_more:
            query = QueryBase(**base_query_dict, limit=page_size, offset=page * page_size)

            output = func(*args, **kwargs, query=query)
            if len(output) == 0: break
            has_more = len(output) == page_size
            res += output
            page += 1
        return res



from ..schemas import QueryBase
from bson import ObjectId
from typing import Optional, List, Union, Tuple
from keble_helpers import is_pydantic_field_empty
from pymongo import ASCENDING, DESCENDING


def cleanse_query_for_first(query: Optional[QueryBase]) -> dict:
    if query is None: return {}
    assert query.filters is None or isinstance(query.filters, dict), "Query.filters only accept dict of Mongo query"
    base = query.filters if query.filters is not None else {}
    base = __build_id_query(base, query)
    return base


def cleanse_query_for_list(query: QueryBase) -> dict:
    base = query.filters if query.filters is not None else {}
    assert query.filters is None or isinstance(query.filters, dict), "Query.filters only accept dict of Mongo query"
    base = __build_ids_query(base, query)
    return base


def cleanse_query_order_by(query: QueryBase) -> Optional[List[Tuple[str, int]]]:
    if query.order_by is not None:
        assert isinstance(query.order_by,
                          list), "Mongodb accept only List[Tuple[str, ASCENDING | DESCENDING]] type ordering"
        if len(query.order_by) == 0: return None
        for item in query.order_by:
            assert isinstance(item, tuple) and len(item) == 2 and isinstance(item[0], str) and item[1] in [DESCENDING,
                                                                                                           ASCENDING], "Mongodb accept only List[Tuple[str, ASCENDING | DESCENDING]] type ordering"
        return query.order_by


MaybeObjectId = Union[str, ObjectId]


def build_object_id(_id: Union[MaybeObjectId, List[MaybeObjectId]]) -> Union[ObjectId, List[ObjectId]]:
    if isinstance(_id, list): return [build_object_id(i) for i in _id]
    if isinstance(_id, str): return ObjectId(_id)
    return _id


def __build_id_query(base_query: dict, query: QueryBase) -> dict:
    if is_pydantic_field_empty(query, 'id'): return base_query
    # use $and if necessary
    q = {'_id': build_object_id(query.id)}
    return {'$and': [base_query, q]} if '_id' in base_query else {**base_query, **q}


def __build_ids_query(base_query: dict, query: QueryBase) -> dict:
    if is_pydantic_field_empty(query, 'ids'): return base_query
    # use $and if necessary
    q = {'_id': {'$in': build_object_id(query.ids)}}
    return {'$and': [base_query, q]} if '_id' in base_query else {**base_query, **q}


def ensure_object_id(id: MaybeObjectId) -> ObjectId:
    if isinstance(id, str): return ObjectId(id)
    return id



from typing import Optional, List, Union, Any
from ..schemas import QueryBase
from sqlmodel import and_
from pydantic import BaseModel, ConfigDict
from sqlalchemy import BinaryExpression


class SqlQueryParsed(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    where: Optional[Union[BinaryExpression, Any]]


def parse_query_for_first(model, query: Optional[QueryBase]) -> SqlQueryParsed:
    where_list = __parse_ids(model, query) + __parse_filters(model, query=query)
    return SqlQueryParsed(where=_build_where(where_list))


def parse_query_for_list(model, query: Optional[QueryBase]) -> SqlQueryParsed:
    where_list = __parse_ids(model, query) + __parse_filters(model, query=query)
    return SqlQueryParsed(where=_build_where(where_list))


def _build_where(where_list: List[BinaryExpression]) -> Optional[BinaryExpression]:
    if len(where_list) == 0:
        return None
    if len(where_list) > 1:
        return and_(*where_list)
    return where_list[0]


def __parse_ids(model, query: Optional[QueryBase]) -> List[BinaryExpression]:
    if query is None: return []
    where: List[BinaryExpression] = []
    if query.id is not None:
        where.append(
            model.id == query.id
        )
    if query.ids is not None:
        where.append(
            model.id.in_(query.ids)
        )
    return where


def __parse_filters(
        model,  # noqa
        query: Optional[QueryBase]) -> List[BinaryExpression]:
    if query is None or query.filters is None or len(query.filters) == 0: return []
    assert not isinstance(query.filters, dict), "Query.filters only accept list of expression for SQL query"
    return query.filters
